/* 
 * Slide show -- Draw in large window with transition effects
 */

#include <tk.h>
#include <stdlib.h>
#include "sldExternals.h"
#include "slide.h"

int Sld_show( slidePtr,transition )
Slide *slidePtr;
char *transition;
{
    Tk_Window tkwin = slidePtr->tkwin;
    Window root = RootWindow( Tk_Display(tkwin), Tk_ScreenNumber(tkwin)); 
	Pixmap dcopy = None;    
	int h,w,rw,rh,zoom,i,j,k;
	unsigned long x,y,x1,y1;
	SldLayer *layerPtr;
    SldItem *itemPtr;
    XWindowAttributes attribs;
    XSetWindowAttributes xsattr;
    Window wshow = slidePtr->wshow;
    unsigned long attrmask = CWOverrideRedirect;  
	GC gc;
	XGCValues gcv;
    int sv_zoom;
            
    xsattr.override_redirect = True;                           
                                                 
    if (strlen(transition) == 0) {
        XDestroyWindow(Tk_Display(tkwin),wshow);
        slidePtr->wshow = (Window)0;
        return TCL_OK;
    }
    w = slidePtr->sWidth;
    h = slidePtr->sHeight;
    XGetWindowAttributes( Tk_Display(tkwin),root,&attribs );
    rw = attribs.width;
    rh = attribs.height;
    zoom = rw*16/w;
    if (rh*16/h < zoom) zoom = rh*16/h;

	dcopy = Tk_GetPixmap(Tk_Display(tkwin), Tk_WindowId(tkwin),
			 rw, rh, DefaultDepthOfScreen(Tk_Screen(tkwin)));
    XFillRectangle(Tk_Display(tkwin),dcopy,slidePtr->gc,0,0,rw,rh);
	Sld_DrawBackground(slidePtr,zoom,dcopy);

    /* save zoom while redrawing */
    sv_zoom = slidePtr->zoom; 
    slidePtr->zoom = zoom;
	for (layerPtr = slidePtr->firstLayer; layerPtr != NULL;
	     layerPtr = layerPtr->nextLayer) {
    	for (itemPtr = layerPtr->firstGluedPtr; itemPtr != NULL;
	         itemPtr = itemPtr->nextItem) {
		    Sld_DisplayItem(slidePtr, dcopy, itemPtr, 0, 0);
	    }
    	for (itemPtr = layerPtr->firstItemPtr; itemPtr != NULL;
	         itemPtr = itemPtr->nextItem) {
		    Sld_DisplayItem(slidePtr, dcopy, itemPtr,0,0); 
	    }
	}
    slidePtr->zoom = sv_zoom;
    
    /* create the slide window if neccessary */
    if (wshow == (Window)0) {
        slidePtr->wshow = wshow = 
        XCreateWindow(Tk_Display(tkwin),root,0,0,rw,rh,0,
            Tk_Depth(tkwin),InputOutput,Tk_Visual(tkwin),attrmask,&xsattr);
        XMapRaised(Tk_Display(tkwin),wshow);
        
    }
    /* check transition desired */
    if (strcmp(transition,"hcurtain")==0) {
        for (i=0;i<rw;i++) {
            XCopyArea(slidePtr->display, dcopy, wshow, slidePtr->gc,
                i,0,1,rh,i,0);
        }
    }
    else if (strcmp(transition,"vcurtain")==0) {
        for (i=0;i<rh;i++) {
            for (j=0;j<rw;j+=8) {
                XCopyArea(slidePtr->display, dcopy, wshow, slidePtr->gc,
                    j,i,8,1,j,i);
            }
        }
    }
    else if (strcmp(transition,"glaze")==0) {
        for (j=0;j<8;j++) {
            for (i=0;i<rh;i+=8) {
                for (k=0;k<rw;k+=8) {
                    XCopyArea(slidePtr->display, dcopy, wshow, slidePtr->gc,
                        k,i+j,8,1,k,i+j);
                }
            }
        }
    }
    else if (strcmp(transition,"lines")==0) {
    	gcv.tile = dcopy;
        gcv.line_width = 3;
    	gcv.fill_style = FillTiled;
    	gc = XCreateGC(slidePtr->display,wshow,
    	    GCLineWidth|GCTile|GCFillStyle,&gcv);
        for (i=0;i<rw*5;i++) {
            x = ((random()/1024)*rw)/(RAND_MAX/1024);
            y = ((random()/1024)*rh)/(RAND_MAX/1024);
            x1 = ((random()/1024)*rw)/(RAND_MAX/1024);
            y1 = ((random()/1024)*rh)/(RAND_MAX/1024);
            XDrawLine(slidePtr->display, wshow, gc,
                x,y,x1,y1);
        }
        XFreeGC(slidePtr->display,gc);
        XCopyArea(slidePtr->display, dcopy, wshow, slidePtr->gc, 0, 0, 
            rw, rh, 0, 0 );
    }
    else if (strcmp(transition,"dots")==0) {
    	gcv.tile = dcopy;
        gcv.line_width = 2;
    	gcv.fill_style = FillTiled;
    	gc = XCreateGC(slidePtr->display,wshow,
    	    GCLineWidth|GCTile|GCFillStyle,&gcv);
        for (i=0;i<rw*rh;i++) {
            x = ((random()/1024)*rw)/(RAND_MAX/1024);
            y = ((random()/1024)*rh)/(RAND_MAX/1024);
            XFillRectangle(slidePtr->display, wshow, gc, x, y, 2, 2);
        }
        XFreeGC(slidePtr->display,gc);
        XCopyArea(slidePtr->display, dcopy, wshow, slidePtr->gc, 0, 0, 
            rw, rh, 0, 0 );
    }
    else if (strcmp(transition,"circular")==0) {
    	gcv.tile = dcopy;
        gcv.line_width = 2;
    	gcv.fill_style = FillTiled;
    	gc = XCreateGC(slidePtr->display,wshow,
    	    GCLineWidth|GCTile|GCFillStyle,&gcv);
        for (i=0;i<360*64;i+=32) {
            XFillArc(slidePtr->display, wshow, gc, -rw/4, -rh/4,
                rw*3/2, rh*3/2, i, 32);
        }
        XFreeGC(slidePtr->display,gc);
        XCopyArea(slidePtr->display, dcopy, wshow, slidePtr->gc, 0, 0, 
            rw, rh, 0, 0 );
    }
    else if (strcmp(transition,"grow")==0) {
    	gcv.tile = dcopy;
        gcv.line_width = 8;
    	gcv.fill_style = FillTiled;
    	gc = XCreateGC(slidePtr->display,wshow,
    	    GCLineWidth|GCTile|GCFillStyle,&gcv);
        for (i=0;i<rw*3/2;i+=(i<rw/8?1:8)) {
            XDrawArc(slidePtr->display, wshow, gc, (rw-i)/2, (rh-i)/2,
                i, i, 0, 360*64);
        }
        XFreeGC(slidePtr->display,gc);
    }
    /* default is to just replace */
    else {
        XCopyArea(slidePtr->display, dcopy, wshow, slidePtr->gc, 0, 0, 
            rw, rh, 0, 0 );
    }
    XFlush(Tk_Display(tkwin));
	Tk_FreePixmap(Tk_Display(tkwin), dcopy);
    return TCL_OK;
}
