/*
 * slide.c --
 */

#include <tk.h>
#include "sldExternals.h"
#include "slide.h"

static void
 appendSelectElement(Tcl_Interp * interp, SldItem * itemPtr, int x, int y);

/*
 * Information used for argv parsing.
 */

static Tk_ConfigSpec configSpecs[] =
{
	{TK_CONFIG_INT, "-width", "width", "Width",
	 "300", Tk_Offset(Slide, width), 0},
	{TK_CONFIG_INT, "-height", "height", "Height",
	 "150", Tk_Offset(Slide, height), 0},
	{TK_CONFIG_INT, "-swidth", "swidth", "Swidth",
	 "800", Tk_Offset(Slide, sWidth), 0},
	{TK_CONFIG_INT, "-sheight", "sheight", "Sheight",
	 "600", Tk_Offset(Slide, sHeight), 0},
    {TK_CONFIG_INT, "-zoom", "zoom", "Zoom",
     "16", Tk_Offset(Slide, zoom), 0},
	{TK_CONFIG_BORDER, "-background", "background", "Background",
	 "#d9d9d9", Tk_Offset(Slide, bgBorder), TK_CONFIG_COLOR_ONLY},
	{TK_CONFIG_BORDER, "-background", "background", "Background",
	 "white", Tk_Offset(Slide, bgBorder), TK_CONFIG_MONO_ONLY},
	{TK_CONFIG_SYNONYM, "-bd", "borderWidth", (char *) NULL,
	 (char *) NULL, 0, 0},
	{TK_CONFIG_SYNONYM, "-bg", "background", (char *) NULL,
	 (char *) NULL, 0, 0},
	{TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
	 "2", Tk_Offset(Slide, borderWidth), 0},
	{TK_CONFIG_SYNONYM, "-fg", "foreground", (char *) NULL,
	 (char *) NULL, 0, 0},
	{TK_CONFIG_BORDER, "-foreground", "foreground", "Foreground",
	 "#b03060", Tk_Offset(Slide, fgBorder), TK_CONFIG_COLOR_ONLY},
	{TK_CONFIG_BORDER, "-foreground", "foreground", "Foreground",
	 "black", Tk_Offset(Slide, fgBorder), TK_CONFIG_MONO_ONLY},
	{TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
	 "raised", Tk_Offset(Slide, relief), 0},
	{TK_CONFIG_INT, "-grid", "grid", "Grid",
	 "0", Tk_Offset(Slide, grid), 0},
	{TK_CONFIG_INT, "-gridsize", "gridsize", "Gridsize",
	 "0", Tk_Offset(Slide, gridsize), 0},
	{TK_CONFIG_COLOR, "-gridcolor", "gridcolor", "Gridcolor",
	 "#000000", Tk_Offset(Slide, gridColor), 0},
	{TK_CONFIG_STRING, "-gradient", "gradient", "Gradient",
	 "", Tk_Offset(Slide, gradient), 0},
	{TK_CONFIG_COLOR, "-gradbegin", "gradbegin", "Gradbegin",
	 "#202020", Tk_Offset(Slide, gradbegin), 0},
	{TK_CONFIG_COLOR, "-gradend", "gradend", "Gradend",
	 "#80c0c0", Tk_Offset(Slide, gradend), 0},
	{TK_CONFIG_COLOR, "-outcolor", "outcolor", "Outcolor",
	 "#404000", Tk_Offset(Slide,outcolor), 0},
    {TK_CONFIG_STRING, "-xscrollcommand","xScrollCommand","ScrollCommand",
     "",Tk_Offset(Slide,xScrollCmd),TK_CONFIG_NULL_OK},
    {TK_CONFIG_STRING, "-yscrollcommand","yScrollCommand","ScrollCommand",
     "",Tk_Offset(Slide,yScrollCmd),TK_CONFIG_NULL_OK},
	{TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	 (char *) NULL, 0, 0}
};

/*
 * Forward declarations for procedures defined later in this file:
 */

int SlideCmd _ANSI_ARGS_((ClientData clientData,
			  Tcl_Interp * interp, int argc, char **argv));
static void SlideCmdDeletedProc _ANSI_ARGS_((
						    ClientData clientData));
static int SlideConfigure _ANSI_ARGS_((Tcl_Interp * interp,
				    Slide * slidePtr, int argc, char **argv,
				       int flags));
static void SlideDestroy _ANSI_ARGS_((char *memPtr));
static void SlideDisplay _ANSI_ARGS_((ClientData clientData));
static void SlideEventProc _ANSI_ARGS_((ClientData clientData,
					XEvent * eventPtr));
static int SlideWidgetCmd _ANSI_ARGS_((ClientData clientData,
				       Tcl_Interp *, int argc, char **argv));

/*
 *--------------------------------------------------------------
 *
 * SlideCmd --
 *
 *  This procedure is invoked to process the "slide" Tcl
 *  command.  It creates a new "slide" widget.
 *
 * Results:
 *  A standard Tcl result.
 *
 * Side effects:
 *  A new widget is created and configured.
 *
 *--------------------------------------------------------------
 */

int SlideCmd(clientData, interp, argc, argv)
ClientData clientData;		/*
				   * Main window associated with
				   * 
				   * * interpreter. 
				 */
Tcl_Interp *interp;		/*
				   * Current interpreter. 
				 */
int argc;			/*
				   * Number of arguments. 
				 */
char **argv;			/*
				   * Argument strings. 
				 */
{
	Tk_Window main = (Tk_Window) clientData;
	Slide *slidePtr;
	Tk_Window tkwin;

	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			   argv[0], " pathName ?options?\"", (char *) NULL);
		return TCL_ERROR;
	}
	tkwin = Tk_CreateWindowFromPath(interp, main, argv[1], (char *) NULL);
	if (tkwin == NULL) {
		return TCL_ERROR;
	}
	Tk_SetClass(tkwin, "Slide");

	/*
	 * Allocate and initialize the widget record.
	 */

	slidePtr = (Slide *) ckalloc(sizeof(Slide));
	slidePtr->tkwin = tkwin;
	slidePtr->display = Tk_Display(tkwin);
	slidePtr->interp = interp;
	slidePtr->widgetCmd = Tcl_CreateCommand(interp,
			       Tk_PathName(slidePtr->tkwin), SlideWidgetCmd,
				(ClientData) slidePtr, SlideCmdDeletedProc);
	slidePtr->currentLayer = NULL;
	slidePtr->firstLayer = NULL;
	slidePtr->lastLayer = NULL;
	slidePtr->nextId = 1;
    slidePtr->wshow = (Window)0;
	slidePtr->rubberband = 0;
	slidePtr->zoom = 16;
	slidePtr->grid = 0;
	slidePtr->gridsize = 0;
	slidePtr->gridColor = NULL;
	slidePtr->gridNpoints = 0;
	slidePtr->gridPoints = NULL;
	slidePtr->backPixmap = 0;
	slidePtr->gradient = "";
	slidePtr->gradbegin = NULL;
	slidePtr->gradend = NULL;
    slidePtr->outcolor = NULL;
	slidePtr->borderWidth = 0;
	slidePtr->width = 500;
	slidePtr->height = 300;
    slidePtr->sWidth = 800;
    slidePtr->sHeight = 600;
    slidePtr->xView = 0;
    slidePtr->yView = 0;
    slidePtr->xScrollCmd = NULL;
    slidePtr->yScrollCmd = NULL;
	slidePtr->bgBorder = NULL;
	slidePtr->fgBorder = NULL;
	slidePtr->relief = TK_RELIEF_FLAT;
	slidePtr->gc = None;
	slidePtr->flags = 0;

	Tk_CreateEventHandler(slidePtr->tkwin,
			      ExposureMask | StructureNotifyMask,
			      SlideEventProc, (ClientData) slidePtr);
	if (SlideConfigure(interp, slidePtr, argc - 2, argv + 2, 0) != TCL_OK) {
		Tk_DestroyWindow(slidePtr->tkwin);
		return TCL_ERROR;
	}
	interp->result = Tk_PathName(slidePtr->tkwin);
	return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * SlideWidgetCmd --
 *
 *  This procedure is invoked to process the Tcl command
 *  that corresponds to a widget managed by this module.
 *  See the user documentation for details on what it does.
 *
 * Results:
 *  A standard Tcl result.
 *
 * Side effects:
 *  See the user documentation.
 *
 *--------------------------------------------------------------
 */

static int SlideWidgetCmd(clientData, interp, argc, argv)
ClientData clientData;		/*
				   * Information about slide widget. 
				 */
Tcl_Interp *interp;		/*
				   * Current interpreter. 
				 */
int argc;			/*
				   * Number of arguments. 
				 */
char **argv;			/*
				   * Argument strings. 
				 */
{
	Slide *slidePtr = (Slide *) clientData;
	int result = TCL_OK;
	size_t length;
	char c;
	int itemId;
	SldItem *itemPtr;
	char buff[5];

	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " option ?arg arg ...?\"", (char *) NULL);
		return TCL_ERROR;
	}
	Tcl_Preserve((ClientData) slidePtr);
	c = argv[1][0];
	length = strlen(argv[1]);
	if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0)
	    && (length >= 2)) {
		if (argc != 3) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
				  argv[0], " cget option\"", (char *) NULL);
			goto error;
		}
		result = Tk_ConfigureValue(interp, slidePtr->tkwin, configSpecs,
					   (char *) slidePtr, argv[2], 0);
	} else if ((c == 'c') && (strncmp(argv[1], "configure", length) == 0)
		   && (length >= 2)) {
		if (argc == 2) {
			result = Tk_ConfigureInfo(interp, slidePtr->tkwin, configSpecs,
				       (char *) slidePtr, (char *) NULL, 0);
		} else if (argc == 3) {
			result = Tk_ConfigureInfo(interp, slidePtr->tkwin, configSpecs,
					     (char *) slidePtr, argv[2], 0);
		} else {
			result = SlideConfigure(interp, slidePtr, argc - 2, argv + 2,
						TK_CONFIG_ARGV_ONLY);
		}
	} else if ((c == 'c') && (strncmp(argv[1], "create", length) == 0)
		   && (length >= 2)) {
		return Sld_CreateItem(slidePtr, argc - 2, argv + 2);
	} else if ((c == 'd') && (strncmp(argv[1], "delete", length) == 0)
		   && (length >= 2)) {
		return Sld_DeleteItem(slidePtr, argc - 2, argv + 2);
	} else if ((c == 'l') && (strncmp(argv[1], "layer", length) == 0)
		   && (length >= 2)) {
		return Sld_LayerCmd(slidePtr, argc - 2, argv + 2);
	} else if ((c == 'i') && (strncmp(argv[1], "itemconfigure", length) == 0)
		   && (length >= 2)) {
		if (argc < 3) {
			Tcl_AppendResult(slidePtr->interp, "wrong # args: ",
			      argv[0], " itemconfig itemId", (char *) NULL);
			return TCL_ERROR;
		}
		if (Tcl_GetInt(slidePtr->interp, argv[2], &itemId) != TCL_OK) {
			Tcl_AppendResult(slidePtr->interp, "wrong item id: ",
			      argv[0], " itemconfig itemId", (char *) NULL);
			return TCL_ERROR;
		}
		return Sld_ConfigureItem(slidePtr, itemId, argc - 3, argv + 3);
	} else if ((c == 'c') && (strncmp(argv[1], "coordinates", length) == 0)
		   && (length >= 2)) {
		if (argc < 3) {
			Tcl_AppendResult(slidePtr->interp, "wrong # args: ",
			     argv[0], " coordinates itemId", (char *) NULL);
			return TCL_ERROR;
		}
		if (Tcl_GetInt(slidePtr->interp, argv[2], &itemId) != TCL_OK) {
			Tcl_AppendResult(slidePtr->interp, "wrong item command: ",
			     argv[0], " coordinates itemId", (char *) NULL);
			return TCL_ERROR;
		}
		return Sld_ItemCoords(slidePtr, itemId, argc - 3, argv + 3);
	} else if ((c == 'n') && (strncmp(argv[1], "names", length) == 0)
		   && (length >= 2)) {
		if (slidePtr->currentLayer == NULL) {
			sprintf(slidePtr->interp->result, "no current layer");
			return TCL_ERROR;
		}
		for (itemPtr = slidePtr->currentLayer->firstItemPtr;
		     itemPtr; itemPtr = itemPtr->nextItem) {
			sprintf(buff, "%d", itemPtr->id);
			Tcl_AppendElement(slidePtr->interp, buff);
		}
		return TCL_OK;
	} else if ((c == 'g') && (strncmp(argv[1], "gnames", length) == 0)
		   && (length >= 2)) {
		if (slidePtr->currentLayer == NULL) {
			sprintf(slidePtr->interp->result, "no current layer");
			return TCL_ERROR;
		}
		for (itemPtr = slidePtr->currentLayer->firstGluedPtr;
		     itemPtr; itemPtr = itemPtr->nextItem) {
			sprintf(buff, "%d", itemPtr->id);
			Tcl_AppendElement(slidePtr->interp, buff);
		}
		return TCL_OK;
	} else if ((c == 'g') && (strncmp(argv[1], "glue", length) == 0)
		   && (length >= 2)) {
		return Sld_GlueItem(slidePtr, argc - 2, argv + 2);
	} else if ((c == 'm') && (strncmp(argv[1], "move", length) == 0)
		   && (length >= 2)) {
		return Sld_MoveItem(slidePtr, argc - 2, argv + 2);
	} else if ((c == 'u') && (strncmp(argv[1], "unglue", length) == 0)
		   && (length >= 2)) {
		return Sld_UnglueItem(slidePtr, argc - 2, argv + 2);
	} else if ((c == 't') && (strncmp(argv[1], "tolayer", length) == 0)
		   && (length >= 2)) {
		return Sld_ToLayerItem(slidePtr, argc - 2, argv + 2);
	} else if ((c == 'r') && (strncmp(argv[1], "raise", length) == 0)
		   && (length >= 2)) {
		return Sld_RaiseItem(slidePtr, argc - 2, argv + 2);
	} else if ((c == 'l') && (strncmp(argv[1], "lower", length) == 0)
		   && (length >= 2)) {
		return Sld_LowerItem(slidePtr, argc - 2, argv + 2);
	} else if ((c == 'r') && (strncmp(argv[1], "rubberband", length) == 0)
		   && (length >= 2)) {
		return Sld_Rubberband(slidePtr, argc - 2, argv + 2);
	} else if ((c == 's') && (strncmp(argv[1], "select", length) == 0)
		   && (length >= 2)) {
		return Sld_SelectItem(slidePtr, argc - 2, argv + 2);
	} else if ((c == 'x') && (strncmp(argv[1], "xview", length) == 0)
		   && (length >= 2)) {
            int count,type;
            double fraction,fraction2;
            
            if (argc == 2) {
                fraction = (double)slidePtr->xView/(double)slidePtr->sWidth;
                fraction2 = (double)(slidePtr->xView+slidePtr->width)
                    /(double)slidePtr->sWidth;
                fraction = FORCE_LIMITS(fraction,0.0,1.0);
                fraction2 = FORCE_LIMITS(fraction2,0.0,1.0);
                sprintf(slidePtr->interp->result,"%g %g",fraction,fraction2);
            }
            else {
                type = Tk_GetScrollInfo(slidePtr->interp, argc, argv,
                    &fraction, &count);
                switch (type) {
                case TK_SCROLL_ERROR:
                    return TCL_ERROR;
                case TK_SCROLL_MOVETO:
                    slidePtr->xView = (int)(fraction * 
                        slidePtr->sWidth + 0.5);
                    break;
                case TK_SCROLL_PAGES:
                    slidePtr->xView = (int)(slidePtr->xView +
                        count*0.9*slidePtr->width);
                    break;
                case TK_SCROLL_UNITS:
                    slidePtr->xView = slidePtr->xView +
                        count*10;
                    break;
                }
                if (slidePtr->xView < -slidePtr->width/2) 
                	slidePtr->xView = -slidePtr->width/2;
                if (slidePtr->xView > (slidePtr->sWidth*slidePtr->zoom/16
                        -slidePtr->width/2))
                    slidePtr->xView = slidePtr->sWidth*slidePtr->zoom/16
                        -slidePtr->width/2;

            	if (!((slidePtr->flags) & WINDOW_REDRAW)) {
            		Tcl_DoWhenIdle(SlideDisplay, (ClientData) slidePtr);
            		slidePtr->flags |= WINDOW_REDRAW;
            	}
            }
            slidePtr->flags |= UPDATE_SCROLLBARS;
	} else if ((c == 'y') && (strncmp(argv[1], "yview", length) == 0)
		   && (length >= 2)) {
            int count,type;
            double fraction,fraction2;
            
            if (argc == 2) {
                fraction = (double)slidePtr->yView/(double)slidePtr->sHeight;
                fraction2 = (double)(slidePtr->yView+slidePtr->height)
                    /(double)slidePtr->sHeight;
                fraction = FORCE_LIMITS(fraction,0.0,1.0);
                fraction2 = FORCE_LIMITS(fraction2,0.0,1.0);
                sprintf(slidePtr->interp->result,"%g %g",fraction,fraction2);
            }
            else {
                type = Tk_GetScrollInfo(slidePtr->interp, argc, argv,
                    &fraction, &count);
                switch (type) {
                case TK_SCROLL_ERROR:
                    return TCL_ERROR;
                case TK_SCROLL_MOVETO:
                    slidePtr->yView = (int)(fraction * 
                        slidePtr->sHeight + 0.5);
                    break;
                case TK_SCROLL_PAGES:
                    slidePtr->yView = (int)(slidePtr->yView +
                        count*0.9*slidePtr->height);
                    break;
                case TK_SCROLL_UNITS:
                    slidePtr->yView = slidePtr->yView +
                        count*10;
                    break;
                }
                if (slidePtr->yView < -slidePtr->height/2) 
                	slidePtr->yView = -slidePtr->height/2;
                if (slidePtr->yView > (slidePtr->sHeight*slidePtr->zoom/16
                        -slidePtr->height/2))
                    slidePtr->yView = slidePtr->sHeight*slidePtr->zoom/16
                        -slidePtr->height/2;

            	if (!((slidePtr->flags) & WINDOW_REDRAW)) {
            		Tcl_DoWhenIdle(SlideDisplay, (ClientData) slidePtr);
            		slidePtr->flags |= WINDOW_REDRAW;
            	}
            }
            slidePtr->flags |= UPDATE_SCROLLBARS;
	} else if ((c == 's') && (strncmp(argv[1], "sx", length) == 0)
		   && (length >= 2)) {
		   int x;
		if (Tcl_GetInt(slidePtr->interp, argv[2], &x) != TCL_OK) {
			Tcl_AppendResult(slidePtr->interp, "wrong point position: ",
			     argv[2], (char *) NULL);
			return TCL_ERROR;
		}
        sprintf(slidePtr->interp->result,"%d",
            (x+slidePtr->xView)*16/slidePtr->zoom);
        return TCL_OK;            
	} else if ((c == 's') && (strncmp(argv[1], "sy", length) == 0)
		   && (length >= 2)) {
		   int y;
		if (Tcl_GetInt(slidePtr->interp, argv[2], &y) != TCL_OK) {
			Tcl_AppendResult(slidePtr->interp, "wrong point position: ",
			     argv[2], (char *) NULL);
			return TCL_ERROR;
		}
        sprintf(slidePtr->interp->result,"%d",
            (y+slidePtr->yView)*16/slidePtr->zoom);
        return TCL_OK;            
	} else if ((c == 'w') && (strncmp(argv[1], "wx", length) == 0)
		   && (length >= 2)) {
		   int x;
		if (Tcl_GetInt(slidePtr->interp, argv[2], &x) != TCL_OK) {
			Tcl_AppendResult(slidePtr->interp, "wrong point position: ",
			     argv[2], (char *) NULL);
			return TCL_ERROR;
		}
        sprintf(slidePtr->interp->result,"%d",(x-slidePtr->xView)*
            slidePtr->zoom/16);
        return TCL_OK;            
	} else if ((c == 'w') && (strncmp(argv[1], "wy", length) == 0)
		   && (length >= 2)) {
		   int y;
		if (Tcl_GetInt(slidePtr->interp, argv[2], &y) != TCL_OK) {
			Tcl_AppendResult(slidePtr->interp, "wrong point position: ",
			     argv[2], (char *) NULL);
			return TCL_ERROR;
		}
        sprintf(slidePtr->interp->result,"%d",(y-slidePtr->yView)*
            slidePtr->zoom/16);
        return TCL_OK;            
	} else if ((c == 't') && (strncmp(argv[1], "tojpeg", length) == 0)
		   && (length >= 2)) {
		if (argc < 2) {
			sprintf(slidePtr->interp->result, "no jpeg filename given");
			return TCL_ERROR;
		}
		return Sld_savejpeg( slidePtr, argv[2] );   
	} else if ((c == 'p') && (strncmp(argv[1], "postscript", length) == 0)
		   && (length >= 2)) {
		if (argc < 2) {
			sprintf(slidePtr->interp->result, "no postscript filename given");
			return TCL_ERROR;
		}
		return Sld_ps_output( slidePtr, argv[2] );   
	} else if ((c == 's') && (strncmp(argv[1], "show", length) == 0)
		   && (length >= 2)) {
		if (argc < 2) {
			Tcl_AppendResult(interp, "wrong # args: must be ",
			   argv[0],argv[1]," transition" );
			return TCL_ERROR;
		}
		return Sld_show( slidePtr, argv[2] );   
	} else {
		Tcl_AppendResult(interp, "bad option \"", argv[1],
		       "\": must be cget, configure, create, delete, layer,"
		         " xview, yview, sx, sy, wx, wy,"
				 " itemconfigure, glue, unglue, tolayer,"
				 " raise, lower, rubberband, select,"
				 " tojpeg, postscript, show,"
				 " move, gnames, or names", (char *) NULL);
		goto error;
	}
	if (!((slidePtr->flags) & WINDOW_REDRAW)) {
		Tcl_DoWhenIdle(SlideDisplay, (ClientData) slidePtr);
		slidePtr->flags |= WINDOW_REDRAW;
	}
	Tcl_Release((ClientData) slidePtr);
	return result;

      error:
	Tcl_Release((ClientData) slidePtr);
	return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * SlideConfigure --
 *
 *  This procedure is called to process an argv/argc list in
 *  conjunction with the Tk option database to configure (or
 *  reconfigure) a slide widget.
 *
 * Results:
 *  The return value is a standard Tcl result.  If TCL_ERROR is
 *  returned, then interp->result contains an error message.
 *
 * Side effects:
 *  Configuration information, such as colors, border width,
 *  etc. get set for slidePtr;  old resources get freed,
 *  if there were any.
 *
 *----------------------------------------------------------------------
 */

static int SlideConfigure(interp, slidePtr, argc, argv, flags)
Tcl_Interp *interp;		/*
				   * Used for error reporting. 
				 */
Slide *slidePtr;		/*
				   * Information about widget. 
				 */
int argc;			/*
				   * Number of valid entries in argv. 
				 */
char **argv;			/*
				   * Arguments. 
				 */
int flags;			/*
				   * Flags to pass to
				   * 
				   * * Tk_ConfigureWidget. 
				 */
{
	Tk_Window tkwin = slidePtr->tkwin;
    int old_zoom = slidePtr->zoom;

	if (Tk_ConfigureWidget(interp, slidePtr->tkwin, configSpecs,
			  argc, argv, (char *) slidePtr, flags) != TCL_OK) {
		return TCL_ERROR;
	}
	/*
	 * Set the background for the window and create a graphics context
	 * for use during redisplay.
	 */

	Tk_SetWindowBackground(slidePtr->tkwin,
			       Tk_3DBorderColor(slidePtr->bgBorder)->pixel);
	if (slidePtr->gc == None) {
		XGCValues gcValues;

		gcValues.function = GXcopy;
		gcValues.graphics_exposures = False;
		slidePtr->gc = Tk_GetGC(slidePtr->tkwin,
			       GCFunction | GCGraphicsExposures, &gcValues);
	}

	if (!((slidePtr->flags) & BACKGROUND_RECONFIG))
		slidePtr->flags |= BACKGROUND_RECONFIG;
	if ((slidePtr->grid) && !((slidePtr->flags) & GRID_RECONFIG))
		slidePtr->flags |= GRID_RECONFIG;

	/*
	 * Register the desired geometry for the window.  Then arrange for
	 * the window to be redisplayed.
	 */

	Tk_GeometryRequest(slidePtr->tkwin, slidePtr->width, slidePtr->height);
	Tk_SetInternalBorder(slidePtr->tkwin, slidePtr->borderWidth);

	if (slidePtr->zoom < 1)
		slidePtr->zoom = 1;
    if (slidePtr->zoom != old_zoom) {
        slidePtr->flags |= CHANGE_ZOOM;
/*    	if (!((slidePtr->flags) & BACKGROUND_RECONFIG))
	    	slidePtr->flags |= BACKGROUND_RECONFIG;
*/
    }
    else {
        slidePtr->flags &= ~CHANGE_ZOOM;
    }

	if (!((slidePtr->flags) & WINDOW_REDRAW)) {
		Tcl_DoWhenIdle(SlideDisplay, (ClientData) slidePtr);
		slidePtr->flags |= WINDOW_REDRAW;
	}
            slidePtr->flags |= UPDATE_SCROLLBARS;
	return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * SlideEventProc --
 *
 *  This procedure is invoked by the Tk dispatcher for various
 *  events on slides.
 *
 * Results:
 *  None.
 *
 * Side effects:
 *  When the window gets deleted, internal structures get
 *  cleaned up.  When it gets exposed, it is redisplayed.
 *
 *--------------------------------------------------------------
 */

static void SlideEventProc(clientData, eventPtr)
ClientData clientData;		/*
				   * Information about window. 
				 */
XEvent *eventPtr;		/*
				   * Information about event. 
				 */
{
	Slide *slidePtr = (Slide *) clientData;

	if (eventPtr->type == Expose) {
		if (!((slidePtr->flags) & WINDOW_REDRAW)) {
			Tcl_DoWhenIdle(SlideDisplay, (ClientData) slidePtr);
			slidePtr->flags |= WINDOW_REDRAW;
		}
	} else if (eventPtr->type == ConfigureNotify) {
    	if (!((slidePtr->flags) & BACKGROUND_RECONFIG))
	    	slidePtr->flags |= BACKGROUND_RECONFIG;

		if (!((slidePtr->flags) & WINDOW_REDRAW)) {
			Tcl_DoWhenIdle(SlideDisplay, (ClientData) slidePtr);
			slidePtr->flags |= WINDOW_REDRAW;
		}
	} else if (eventPtr->type == DestroyNotify) {
		if (slidePtr->tkwin != NULL) {
			slidePtr->tkwin = NULL;
			Tcl_DeleteCommandFromToken(slidePtr->interp,
						   slidePtr->widgetCmd);
		}
		if ((slidePtr->flags) & WINDOW_REDRAW) {
			Tcl_CancelIdleCall(SlideDisplay, (ClientData) slidePtr);
		}
		Tcl_EventuallyFree((ClientData) slidePtr, SlideDestroy);
	}
}

/*
 *----------------------------------------------------------------------
 *
 * SlideCmdDeletedProc --
 *
 *  This procedure is invoked when a widget command is deleted.  If
 *  the widget isn't already in the process of being destroyed,
 *  this command destroys it.
 *
 * Results:
 *  None.
 *
 * Side effects:
 *  The widget is destroyed.
 *
 *----------------------------------------------------------------------
 */

static void SlideCmdDeletedProc(clientData)
ClientData clientData;		/*
				   * Pointer to widget record for widget. 
				 */
{
	Slide *slidePtr = (Slide *) clientData;
	Tk_Window tkwin = slidePtr->tkwin;

	/*
	 * This procedure could be invoked either because the window was
	 * destroyed and the command was then deleted (in which case tkwin
	 * is NULL) or because the command was deleted, and then this procedure
	 * destroys the widget.
	 */

	if (tkwin != NULL) {
		slidePtr->tkwin = NULL;
		Tk_DestroyWindow(tkwin);
	}
}

/*
 *--------------------------------------------------------------
 *
 * SlideDisplay --
 *
 *  This procedure redraws the contents of a slide window.
 *  It is invoked as a do-when-idle handler, so it only runs
 *  when there's nothing else for the application to do.
 *
 * Results:
 *  None.
 *
 * Side effects:
 *  Information appears on the screen.
 *
 *--------------------------------------------------------------
 */

static void SlideDisplay(clientData)
ClientData clientData;		/*
				   * Information about window. 
				 */
{
	Slide *slidePtr = (Slide *) clientData;
	Tk_Window tkwin = slidePtr->tkwin;
	Pixmap d = None;
	GC gc;
	XGCValues gcv;
	int x, c;
	SldLayer *layerPtr;
	int zoom=slidePtr->zoom;

	slidePtr->flags &= ~WINDOW_REDRAW;
	if (!Tk_IsMapped(tkwin)) {
		return;
	}
	/*
	 * Create a pixmap for double-buffering
	 */

	d = Tk_GetPixmap(Tk_Display(tkwin), Tk_WindowId(tkwin),
			 slidePtr->width, slidePtr->height,
			 DefaultDepthOfScreen(Tk_Screen(tkwin)));
	gcv.foreground = slidePtr->outcolor->pixel;
	gcv.function = GXcopy;
	gc = Tk_GetGC(tkwin,GCFunction|GCForeground, &gcv);
	XFillRectangle(Tk_Display(tkwin), d, gc, 0, 0, 
		slidePtr->width, slidePtr->height ); 
	Tk_FreeGC(Tk_Display(tkwin), gc);
	
	/*
	 * Show backPixmap if requested
	 */
	if ((slidePtr->flags) & BACKGROUND_RECONFIG) {
    	if (slidePtr->backPixmap)
	    	Tk_FreePixmap(Tk_Display(tkwin), slidePtr->backPixmap);
	    slidePtr->backPixmap =  Tk_GetPixmap(Tk_Display(tkwin), 
	        Tk_WindowId(tkwin), 
	        slidePtr->sWidth*slidePtr->zoom/16, 
	        slidePtr->sHeight*slidePtr->zoom/16, 
	        DefaultDepthOfScreen(Tk_Screen(tkwin)));
        
		Sld_DrawBackground(slidePtr,slidePtr->zoom,slidePtr->backPixmap);
		/*
		 * redraw glued items in all layers 
		 */
		for (layerPtr = slidePtr->firstLayer; layerPtr != NULL;
		     layerPtr = layerPtr->nextLayer) {
			if (layerPtr->visible)
				Sld_DisplayGluedLayer(slidePtr, slidePtr->backPixmap,
						      layerPtr);
		}
		slidePtr->flags &= ~BACKGROUND_RECONFIG;
	}
	XCopyArea(Tk_Display(tkwin), slidePtr->backPixmap, d,
		  slidePtr->gc, slidePtr->xView, slidePtr->yView, 
		  slidePtr->width, slidePtr->height, 0, 0);

	/*
	 * make new grid points
	 */
	if ((slidePtr->flags) & GRID_RECONFIG) {
		int nx, ny, x, y;
		XPoint *pPtr;
		XRectangle *rPtr;

		if (slidePtr->gridPoints != NULL)
			ckfree(slidePtr->gridPoints);
		if (slidePtr->grid < 50)
			slidePtr->grid = 50;
		nx = (slidePtr->width * slidePtr->zoom) / slidePtr->grid + 1;
		ny = (slidePtr->height * slidePtr->zoom) / slidePtr->grid + 1;
		slidePtr->gridNpoints = nx * ny;
		if (slidePtr->gridsize == 0) {
			slidePtr->gridPoints = (XPoint *) ckalloc(nx * ny * sizeof(XPoint));
			pPtr = slidePtr->gridPoints;
			for (y = 0; y < ny; y++) {
				for (x = 0; x < nx; x++) {
					pPtr->x = (x * slidePtr->grid * 
					    slidePtr->zoom/16 + 5) / 16;
					pPtr->y = (y * slidePtr->grid *
					    slidePtr->zoom/16 + 5) / 16;
					pPtr++;
				}
			}
		} else {
			slidePtr->gridPoints = (XPoint *) ckalloc(nx * ny * sizeof(XRectangle));
			rPtr = (XRectangle *) slidePtr->gridPoints;
			for (y = 0; y < ny; y++) {
				for (x = 0; x < nx; x++) {
					rPtr->x = (x * slidePtr->grid*slidePtr->zoom/16 
					    + 5 - slidePtr->gridsize*slidePtr->zoom/16) / 16;
					rPtr->y = (y * slidePtr->grid*slidePtr->zoom/16 
					    + 5 - slidePtr->gridsize*slidePtr->zoom/16) / 16;
					rPtr->width = slidePtr->gridsize*slidePtr->zoom/16;
					rPtr->height = slidePtr->gridsize*slidePtr->zoom/16;
					rPtr++;
				}
			}
		}
		slidePtr->flags &= ~GRID_RECONFIG;
	} else if ((slidePtr->grid == 0) && (slidePtr->gridPoints != NULL))
		ckfree(slidePtr->gridPoints);
	/*
	 * display grid
	 */
	if (slidePtr->grid && (slidePtr->gridPoints != NULL)) {
		gcv.foreground = slidePtr->gridColor->pixel;
		gc = Tk_GetGC(tkwin, GCForeground, &gcv);
		if (slidePtr->gridsize == 0)
			XDrawPoints(Tk_Display(tkwin), d, gc, slidePtr->gridPoints,
				    slidePtr->gridNpoints, CoordModeOrigin);
		else
			XDrawRectangles(Tk_Display(tkwin), d, gc,
					(XRectangle *) slidePtr->gridPoints,
					slidePtr->gridNpoints);

		Tk_FreeGC(Tk_Display(tkwin), gc);
	}
	/*
	 * show all enabled layers
	 */
	for (layerPtr = slidePtr->firstLayer; layerPtr != NULL;
	     layerPtr = layerPtr->nextLayer) {
		if (layerPtr->visible)
			Sld_DisplayLayer(slidePtr, d, layerPtr);
	}

	XCopyArea(Tk_Display(tkwin), d, Tk_WindowId(tkwin), slidePtr->gc,
	      0, 0, slidePtr->width, slidePtr->height, 0, 0);
	Tk_FreePixmap(Tk_Display(tkwin), d);
	if (slidePtr->flags & UPDATE_SCROLLBARS) {
        slidePtr->flags &= ~UPDATE_SCROLLBARS;
        Sld_SetScrollbars( slidePtr );
	}
}


/*
 *----------------------------------------------------------------------
 *
 * SlideDestroy --
 *
 *  This procedure is invoked by Tcl_EventuallyFree or Tcl_Release
 *  to clean up the internal structure of a slide at a safe time
 *  (when no-one is using it anymore).
 *
 * Results:
 *  None.
 *
 * Side effects:
 *  Everything associated with the slide is freed up.
 *
 *----------------------------------------------------------------------
 */

static void SlideDestroy(memPtr)
char *memPtr;			/*
				   * Info about slide widget. 
				 */
{
	Slide *slidePtr = (Slide *) memPtr;
	SldLayer *layerPtr;
	SldItem *itemPtr, *nextItemPtr;

	/* first destroy all items in layers */
	for (layerPtr = slidePtr->firstLayer; layerPtr;
	     layerPtr = layerPtr->nextLayer) {
		/*
		 * remove all items from this layer 
		 */
		itemPtr = layerPtr->firstItemPtr;
		while (itemPtr != NULL) {
			nextItemPtr = itemPtr->nextItem;
			if (itemPtr->kind == SLD_ITEM_IMAGE) {
				XDestroyImage(itemPtr->xi);
			}
			if (itemPtr->kind == SLD_ITEM_TEXT)
				Tk_FreeFont(itemPtr->tkfont);
			ckfree(itemPtr);
			itemPtr = nextItemPtr;
		}
		itemPtr = layerPtr->firstGluedPtr;
		while (itemPtr != NULL) {
			nextItemPtr = itemPtr->nextItem;
			if (itemPtr->kind == SLD_ITEM_IMAGE) {
				XDestroyImage(itemPtr->xi);
			}
			if (itemPtr->kind == SLD_ITEM_TEXT)
				Tk_FreeFont(itemPtr->tkfont);
			ckfree(itemPtr);
			itemPtr = nextItemPtr;
		}
	}

	Tk_FreeOptions(configSpecs, (char *) slidePtr, slidePtr->display, 0);
	if (slidePtr->gc != None) {
		Tk_FreeGC(slidePtr->display, slidePtr->gc);
	}
	ckfree((char *) slidePtr);
}


/*
 *--------------------------------------------------------------
 *
 * Sld_DeleteLayers -- 
 *
 * This procedure delete all objects in a layer and the
 * layer itself
 *
 *--------------------------------------------------------------
 */
int Sld_DeleteLayers(slidePtr, argc, argv)
Slide *slidePtr;
int argc;
char **argv;
{
	int i;
	int layerId, itemId;
	SldLayer *layerPtr, *prevLayerPtr;
	SldItem *itemPtr, *nextItemPtr;

	if (argc < 1) {
		Tcl_AppendResult(slidePtr->interp, "no layer id given", (char *) NULL);
		return TCL_ERROR;
	}
	for (i = 0; i < argc; i++) {
		/*
		 * get layer id to delete 
		 */
		if (Tcl_GetInt(slidePtr->interp, argv[i], &layerId) != TCL_OK) {
			Tcl_AppendElement(slidePtr->interp, "wrong layer id");
			return TCL_ERROR;
		}
		/*
		 * locate the layer in layer's list 
		 */
		layerPtr = slidePtr->firstLayer;
		prevLayerPtr = NULL;
		while (layerPtr != NULL) {
			if (layerPtr->id == layerId)
				break;
			prevLayerPtr = layerPtr;
			layerPtr = layerPtr->nextLayer;
		}
		if (layerPtr == NULL) {
			Tcl_AppendElement(slidePtr->interp, "layer not found");
			return TCL_ERROR;
		}
		/*
		 * remove all items from this layer 
		 */
		itemPtr = layerPtr->firstItemPtr;
		while (itemPtr != NULL) {
			nextItemPtr = itemPtr->nextItem;
			if (itemPtr->kind == SLD_ITEM_IMAGE) {
				XDestroyImage(itemPtr->xi);
			}
			ckfree(itemPtr);
			itemPtr = nextItemPtr;
		}
		/*
		 * if this was the first layer, make it's next the first one 
		 */
		if (prevLayerPtr == NULL) {
			slidePtr->firstLayer = layerPtr->nextLayer;
		} else {
			prevLayerPtr->nextLayer = layerPtr->nextLayer;
		}
		/*
		 * we can't lose the last layer pointer too 
		 */
		if (slidePtr->lastLayer == layerPtr)
			slidePtr->lastLayer = prevLayerPtr;
		/*
		 * if deleting current layer, make current unavailable 
		 */
		if (slidePtr->currentLayer == layerPtr)
			slidePtr->currentLayer = NULL;
		ckfree(layerPtr);
	}
}

static Tk_ConfigSpec layerConfigSpecs[] =
{
	{TK_CONFIG_INT, "-visible", "visible", "Visible",
	 "1", Tk_Offset(SldLayer, visible), 0},
	{TK_CONFIG_INT, "-transparent", "transparent", "Transparent",
	 "0", Tk_Offset(SldLayer, transparent), 0},
	{TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	 (char *) NULL, 0, 0}
};

/*
 *--------------------------------------------------------------
 *
 * Sld_LayerCmd -- 
 *
 * This procedure implements several commands for manipulating 
 * layers
 *
 *--------------------------------------------------------------
 */
int Sld_LayerCmd(slidePtr, argc, argv)
Slide *slidePtr;
int argc;
char **argv;
{
	int result = TCL_OK;
	size_t length;
	char c, buff[5];
	SldLayer *layerPtr;
	int id;

	if (argc < 1) {
		Tcl_AppendResult(slidePtr->interp, "unknown layer subcommand: ",
		"\"new, delete, config, current or names\"", (char *) NULL);
		return TCL_ERROR;
	}
	Tcl_Preserve((ClientData) slidePtr);
	c = argv[0][0];
	length = strlen(argv[0]);
	if ((c == 'n') && (strncmp(argv[0], "new", length) == 0)
	    && (length >= 2)) {
		layerPtr = (SldLayer *) ckalloc(sizeof(SldLayer));
		layerPtr->id = slidePtr->nextId++;
		layerPtr->visible = 1;
		layerPtr->transparent = 0;
		layerPtr->nextLayer = slidePtr->firstLayer;
		if (slidePtr->lastLayer == NULL)
			slidePtr->lastLayer = layerPtr;
		slidePtr->firstLayer = layerPtr;
		layerPtr->nextId = 1;
		layerPtr->firstItemPtr = NULL;
		layerPtr->lastItemPtr = NULL;
		layerPtr->firstGluedPtr = NULL;
		layerPtr->lastGluedPtr = NULL;
		slidePtr->currentLayer = layerPtr;

		sprintf(slidePtr->interp->result, "%d", layerPtr->id);
		return TCL_OK;
	}
	if ((c == 'n') && (strncmp(argv[0], "names", length) == 0)
	    && (length >= 2)) {
		for (layerPtr = slidePtr->firstLayer; layerPtr;
		     layerPtr = layerPtr->nextLayer) {
			sprintf(buff, "%d", layerPtr->id);
			Tcl_AppendElement(slidePtr->interp, buff);
		}
		return TCL_OK;
	}
	if ((c == 'd') && (strncmp(argv[0], "delete", length) == 0)
	    && (length >= 2)) {
		if (argc < 2) {
			Tcl_AppendResult(slidePtr->interp, "wrong # args: should be \"",
			   argv[0], " pathName ?options?\"", (char *) NULL);
			return TCL_ERROR;
		}
		return Sld_DeleteLayers(slidePtr, argc - 1, argv + 1);
	}
	if ((c == 'c') && (strncmp(argv[0], "current", length) == 0)
	    && (length >= 2)) {
		if (argc == 1) {
			if (slidePtr->currentLayer != NULL) {
				sprintf(slidePtr->interp->result, "%d",
					slidePtr->currentLayer->id);
			}
		} else {
			if (Tcl_GetInt(slidePtr->interp, argv[1], &id) != TCL_OK) {
				Tcl_AppendElement(slidePtr->interp, "wrong layer id");
				return TCL_ERROR;
			}
			layerPtr = slidePtr->firstLayer;
			while (layerPtr != NULL) {
				if (layerPtr->id == id)
					break;
				layerPtr = layerPtr->nextLayer;
			}
			if (layerPtr == NULL) {
				Tcl_AppendElement(slidePtr->interp, "layer not found");
				return TCL_ERROR;
			}
			slidePtr->currentLayer = layerPtr;
		}
		return TCL_OK;
	}
	if ((c == 'c') && (strncmp(argv[0], "configure", length) == 0)
	    && (length >= 2)) {
		if (slidePtr->currentLayer == NULL) {
			sprintf(slidePtr->interp->result, "no current layer");
			return TCL_ERROR;
		}
		if (argc == 1) {
			result = Tk_ConfigureInfo(slidePtr->interp, slidePtr->tkwin,
			  layerConfigSpecs, (char *) slidePtr->currentLayer,
						  (char *) NULL, 0);
		} else if (argc == 2) {
			result = Tk_ConfigureInfo(slidePtr->interp, slidePtr->tkwin,
			  layerConfigSpecs, (char *) slidePtr->currentLayer,
						  argv[1], 0);
		} else {
			result = Tk_ConfigureWidget(slidePtr->interp, slidePtr->tkwin,
				       layerConfigSpecs, argc - 1, argv + 1,
						    (char *) slidePtr->currentLayer, TK_CONFIG_ARGV_ONLY);
			if (!((slidePtr->flags) & WINDOW_REDRAW)) {
				Tcl_DoWhenIdle(SlideDisplay, (ClientData) slidePtr);
				slidePtr->flags |= WINDOW_REDRAW;
			}
		}
		return result;
	}
	Tcl_AppendElement(slidePtr->interp, "unknown layer command");
	return TCL_ERROR;
}

void Sld_DisplayLayer(slidePtr, d, layerPtr)
Slide *slidePtr;
Drawable d;
SldLayer *layerPtr;
{
	SldItem *itemPtr;

	for (itemPtr = layerPtr->firstItemPtr; itemPtr != NULL;
	     itemPtr = itemPtr->nextItem) {
		Sld_DisplayItem(slidePtr, d, itemPtr, 
		    slidePtr->xView, slidePtr->yView);
	}
}

void Sld_DisplayGluedLayer(slidePtr, d, layerPtr)
Slide *slidePtr;
Drawable d;
SldLayer *layerPtr;
{
	SldItem *itemPtr;

	for (itemPtr = layerPtr->firstGluedPtr; itemPtr != NULL;
	     itemPtr = itemPtr->nextItem) {
		Sld_DisplayItem(slidePtr, d, itemPtr, 0, 0);
	}
}

static Tk_ConfigSpec itemConfigSpecs[] =
{
	{TK_CONFIG_COLOR, "-fill", "fill", "Fill",
	 "#000000", Tk_Offset(SldItem, color),SLD_LINE_MASK|SLD_TEXT_MASK},
	{TK_CONFIG_COLOR, "-fill", "fill", "Fill",
	 (char *)NULL, Tk_Offset(SldItem, fillColor), SLD_ARC_MASK|
	  SLD_RECTANGLE_MASK|TK_CONFIG_NULL_OK},
	{TK_CONFIG_COLOR, "-outline", "outline", "Outline",
	 "#000000", Tk_Offset(SldItem, color), SLD_ARC_MASK|SLD_RECTANGLE_MASK},
	{TK_CONFIG_INT, "-width", "width", "Width",
	 "1", Tk_Offset(SldItem, width), SLD_LINE_MASK|SLD_RECTANGLE_MASK|
	    SLD_ARC_MASK},
	{TK_CONFIG_INT, "-angle1", "angle1", "Angle1",
	 "0", Tk_Offset(SldItem, angle1), SLD_ARC_MASK},
	{TK_CONFIG_INT, "-angle2", "angle2", "Angle2",
	 "23040", Tk_Offset(SldItem, angle2), SLD_ARC_MASK},
	{TK_CONFIG_STRING, "-image", "image", "Image",
	 "", Tk_Offset(SldItem, image), SLD_IMAGE_MASK},
	{TK_CONFIG_STRING, "-text", "text", "Text",
	 "", Tk_Offset(SldItem, text), SLD_TEXT_MASK},
	{TK_CONFIG_INT, "-rotate", "rotate", "Rotate",
	 "0", Tk_Offset(SldItem, rotate), SLD_TEXT_MASK},
	{TK_CONFIG_FONT, "-font", "font", "Font",
	 "Courier 12", Tk_Offset(SldItem, tkfont), SLD_TEXT_MASK},
	{TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	 (char *) NULL, 0, 0}
};

int configFlags[] =
{SLD_ALL_MASK, SLD_LINE_MASK, SLD_RECTANGLE_MASK,
 SLD_ARC_MASK, SLD_OVAL_MASK, SLD_IMAGE_MASK, SLD_TEXT_MASK};

void Sld_AppendItem(layerPtr, itemPtr)
SldLayer *layerPtr;
SldItem *itemPtr;
{
	if (layerPtr->lastItemPtr)
		layerPtr->lastItemPtr->nextItem = itemPtr;
	layerPtr->lastItemPtr = itemPtr;
	if (layerPtr->firstItemPtr == NULL)
		layerPtr->firstItemPtr = itemPtr;
}

/*
 *--------------------------------------------------------------
 *
 * Sld_CreateItem --
 *
 * This procedure creates a new item in the current layer of
 * the slide widget, after parsing the type of item and its
 * position.
 *
 *--------------------------------------------------------------
 */
int Sld_CreateItem(slidePtr, argc, argv)
Slide *slidePtr;
int argc;
char **argv;
{
	SldLayer *layerPtr = slidePtr->currentLayer;
	SldItem *itemPtr;

	if (argc < 5) {
		Tcl_AppendResult(slidePtr->interp, "wrong # args", (char *) NULL);
		return TCL_ERROR;
	}
	if (layerPtr == NULL) {
		sprintf(slidePtr->interp->result, "no current layer");
		return TCL_ERROR;
	}
	itemPtr = (SldItem *) ckalloc(sizeof(SldItem));
	itemPtr->nextItem = NULL;
	itemPtr->id = layerPtr->nextId++;
	itemPtr->width = 1;
	itemPtr->color = NULL;
	itemPtr->fillColor = NULL;
	itemPtr->xi = NULL;
	itemPtr->image = "";
	itemPtr->angle1 = 0;
	itemPtr->angle2 = 360 * 64;
	itemPtr->text = NULL;
	itemPtr->rotate = 0;
	itemPtr->tkfont = NULL;

	if (Tcl_GetInt(slidePtr->interp, argv[1], &itemPtr->x1) != TCL_OK) {
		goto error;
	}
	if (Tcl_GetInt(slidePtr->interp, argv[2], &itemPtr->y1) != TCL_OK) {
		goto error;
	}
	if (Tcl_GetInt(slidePtr->interp, argv[3], &itemPtr->x2) != TCL_OK) {
		goto error;
	}
	if (Tcl_GetInt(slidePtr->interp, argv[4], &itemPtr->y2) != TCL_OK) {
		goto error;
	}
	if (strcmp(argv[0], "line") == 0) {
		itemPtr->kind = SLD_ITEM_LINE;
	} else if (strcmp(argv[0], "arc") == 0) {
		itemPtr->kind = SLD_ITEM_ARC;
	} else if (strcmp(argv[0], "rectangle") == 0) {
		itemPtr->kind = SLD_ITEM_RECTANGLE;
	} else if (strcmp(argv[0], "image") == 0) {
		itemPtr->kind = SLD_ITEM_IMAGE;
	} else if (strcmp(argv[0], "text") == 0) {
		itemPtr->kind = SLD_ITEM_TEXT;
	} else {
		Tcl_AppendResult(slidePtr->interp, "item kind unknown",
				 (char *) NULL);
		goto error2;
	}
	Sld_AppendItem(layerPtr, itemPtr);

	if (Tk_ConfigureWidget(slidePtr->interp, slidePtr->tkwin,
			       itemConfigSpecs, argc - 5, argv + 5,
		    (char *) itemPtr, configFlags[itemPtr->kind]) != TCL_OK)
		goto error2;

	if (itemPtr->kind == SLD_ITEM_TEXT) {
		Tk_FontMetrics fm;
		Tk_GetFontMetrics(itemPtr->tkfont, &fm);
		itemPtr->x2 = Tk_TextWidth(itemPtr->tkfont, itemPtr->text,
				       strlen(itemPtr->text)) + itemPtr->x1;
		itemPtr->y2 = itemPtr->y1 - fm.ascent;
	}
	if (!((slidePtr->flags) & WINDOW_REDRAW)) {
		Tcl_DoWhenIdle(SlideDisplay, (ClientData) slidePtr);
		slidePtr->flags |= WINDOW_REDRAW;
	}
	sprintf(slidePtr->interp->result, "%d", itemPtr->id);
	return TCL_OK;
      error:
	Tcl_AppendElement(slidePtr->interp, "wrong coords");
      error2:
	ckfree(itemPtr);
	return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * Sld_ConfigureItem -- change item properties
 *
 *--------------------------------------------------------------
 */
int Sld_ConfigureItem(slidePtr, itemId, argc, argv)
Slide *slidePtr;
int itemId;
int argc;
char **argv;
{
	SldItem *itemPtr;
	int result,glued=0;

	if (slidePtr->currentLayer == NULL) {
		sprintf(slidePtr->interp->result, "no current layer");
		return TCL_ERROR;
	}
	/*
	 * locate the item in list for current layer 
	 */
	itemPtr = slidePtr->currentLayer->firstItemPtr;
	while (itemPtr != NULL) {
		if (itemPtr->id == itemId)
			break;
		itemPtr = itemPtr->nextItem;
	}
    /* if not found, try glued items */
	if (itemPtr == NULL) {
	    glued++;
    	itemPtr = slidePtr->currentLayer->firstGluedPtr;
    	while (itemPtr != NULL) {
	    	if (itemPtr->id == itemId)
		    	break;
    		itemPtr = itemPtr->nextItem;
	    }
    }
	if (itemPtr == NULL) {
		sprintf(slidePtr->interp->result, "item %d not found", itemId);
		return TCL_ERROR;
	}
	/*
	 * try to get or set configuration parameters 
	 */
	if (argc == 0) {
		result = Tk_ConfigureInfo(slidePtr->interp, slidePtr->tkwin,
					  itemConfigSpecs, (char *) itemPtr,
				 (char *) NULL, configFlags[itemPtr->kind]);
	} else if (argc == 1) {
		result = Tk_ConfigureInfo(slidePtr->interp, slidePtr->tkwin,
					  itemConfigSpecs, (char *) itemPtr,
				       argv[0], configFlags[itemPtr->kind]);
	} else {
		result = Tk_ConfigureWidget(slidePtr->interp, slidePtr->tkwin,
					    itemConfigSpecs, argc, argv,
				     (char *) itemPtr, TK_CONFIG_ARGV_ONLY);
		if (!((slidePtr->flags) & WINDOW_REDRAW)) {
			Tcl_DoWhenIdle(SlideDisplay, (ClientData) slidePtr);
			slidePtr->flags |= WINDOW_REDRAW;
		}
	}
	if (glued)
	    if (!((slidePtr->flags) & BACKGROUND_RECONFIG))
	    	slidePtr->flags |= BACKGROUND_RECONFIG;
	return result;
}

/*
 *--------------------------------------------------------------
 *
 * Sld_ItemCoords -- get/set item coordinates
 *
 *--------------------------------------------------------------
 */
int Sld_ItemCoords(slidePtr, itemId, argc, argv)
Slide *slidePtr;
int itemId;
int argc;
char **argv;
{
	SldItem *itemPtr;
	int i,glued;

	if (slidePtr->currentLayer == NULL) {
		sprintf(slidePtr->interp->result, "no current layer");
		return TCL_ERROR;
	}
	/*
	 * locate the item in list for current layer 
	 */
	itemPtr = slidePtr->currentLayer->firstItemPtr;
	while (itemPtr != NULL) {
		if (itemPtr->id == itemId)
			break;
		itemPtr = itemPtr->nextItem;
	}
    /* if not found, try glued items */
	if (itemPtr == NULL) {
	    glued++;
    	itemPtr = slidePtr->currentLayer->firstGluedPtr;
    	while (itemPtr != NULL) {
	    	if (itemPtr->id == itemId)
		    	break;
    		itemPtr = itemPtr->nextItem;
	    }
    }
	if (itemPtr == NULL) {
		sprintf(slidePtr->interp->result, "item %d not found", itemId);
		return TCL_ERROR;
	}
	/*
	 * try to get or set coordinates 
	 */
	if (argc == 0) {
		sprintf(slidePtr->interp->result, "%d %d %d %d",
			itemPtr->x1, itemPtr->y1, itemPtr->x2, itemPtr->y2);
		return TCL_OK;
	} else if (argc != 4) {
		Tcl_AppendResult(slidePtr->interp, "wrong # coordinates, must be:",
				 "\" x1 y1 x2 y2 \"", (char *) NULL);
		return TCL_ERROR;
	} else {
		if (Tcl_GetInt(slidePtr->interp, argv[0], &itemPtr->x1) != TCL_OK) {
			goto error;
		}
		if (Tcl_GetInt(slidePtr->interp, argv[1], &itemPtr->y1) != TCL_OK) {
			goto error;
		}
		if (Tcl_GetInt(slidePtr->interp, argv[2], &itemPtr->x2) != TCL_OK) {
			goto error;
		}
		if (Tcl_GetInt(slidePtr->interp, argv[3], &itemPtr->y2) != TCL_OK) {
			goto error;
		}
		if (!((slidePtr->flags) & WINDOW_REDRAW)) {
			Tcl_DoWhenIdle(SlideDisplay, (ClientData) slidePtr);
			slidePtr->flags |= WINDOW_REDRAW;
		}
    	if (glued)
	        if (!((slidePtr->flags) & BACKGROUND_RECONFIG))
	        	slidePtr->flags |= BACKGROUND_RECONFIG;
	}
	return TCL_OK;
      error:
	Tcl_AppendResult(slidePtr->interp, "wrong # coordinates, must be:",
			 "\" x1 y1 x2 y2 \"", (char *) NULL);
	return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * Sld_DisplayItem -- actually draw items 
 *
 *--------------------------------------------------------------
 */
void Sld_DisplayItem(slidePtr, d, itemPtr, xoff, yoff)
Slide *slidePtr;
Drawable d;
SldItem *itemPtr;
int xoff,yoff;
{
	GC gc,gc1;
	XGCValues gcv;
	int zoom=slidePtr->zoom;

    if ((itemPtr->kind != SLD_ITEM_IMAGE)&&
        (itemPtr->kind != SLD_ITEM_TEXT)) {
    	gcv.foreground = itemPtr->color->pixel;
	    gcv.line_width = itemPtr->width*zoom/16;
	    gc = Tk_GetGC(slidePtr->tkwin, GCForeground | GCLineWidth, &gcv);
	}
	if (itemPtr->kind == SLD_ITEM_LINE) {
		XDrawLine(slidePtr->display, d, gc, 
		    itemPtr->x1*zoom/16-xoff, itemPtr->y1*zoom/16-yoff, 
		    itemPtr->x2*zoom/16-xoff, itemPtr->y2*zoom/16-yoff);
	} else if (itemPtr->kind == SLD_ITEM_ARC) {
		XDrawArc(slidePtr->display, d, gc, 
		    itemPtr->x1*zoom/16-xoff, itemPtr->y1*zoom/16-yoff,
		    (itemPtr->x2 - itemPtr->x1)*zoom/16, 
		    (itemPtr->y2 - itemPtr->y1)*zoom/16,
		    itemPtr->angle1, itemPtr->angle2);
	} else if (itemPtr->kind == SLD_ITEM_RECTANGLE) {
        if (itemPtr->fillColor) {
        	gcv.foreground = itemPtr->fillColor->pixel;
	        gcv.line_width = itemPtr->width*zoom/16;
	        gc1 = Tk_GetGC(slidePtr->tkwin, GCForeground | GCLineWidth, &gcv);
		    XFillRectangle(slidePtr->display, d, gc1,
		        itemPtr->x1*zoom/16-xoff, itemPtr->y1*zoom/16-yoff, 
		        (itemPtr->x2 - itemPtr->x1)*zoom/16, 
		        (itemPtr->y2 - itemPtr->y1)*zoom/16);
        	Tk_FreeGC(slidePtr->display, gc1);
        }
		XDrawRectangle(slidePtr->display, d, gc, 
		   itemPtr->x1*zoom/16-xoff, itemPtr->y1*zoom/16-yoff,
		  (itemPtr->x2 - itemPtr->x1)*zoom/16, 
		  (itemPtr->y2 - itemPtr->y1)*zoom/16);
	} else if (itemPtr->kind == SLD_ITEM_IMAGE) {
        XImage *xi;
	    gc = Tk_GetGC(slidePtr->tkwin, 0, &gcv);
		if (itemPtr->xi == NULL)
		    Sld_loadjpg(slidePtr, itemPtr);
        if (slidePtr->zoom == 16) {
		    XPutImage(slidePtr->display, d, gc, itemPtr->xi, 0, 0,
		      itemPtr->x1-xoff, itemPtr->y1-yoff, 
		      (itemPtr->x2 - itemPtr->x1),
			  (itemPtr->y2 - itemPtr->y1));
        }
        else {
            xi = Sld_zoom_image( slidePtr->tkwin, itemPtr->xi, 
                slidePtr->zoom ); 
		    XPutImage(slidePtr->display, d, gc, xi, 0, 0,
		      itemPtr->x1*zoom/16-xoff, itemPtr->y1*zoom/16-yoff, 
		      (itemPtr->x2 - itemPtr->x1)*zoom/16,
			  (itemPtr->y2 - itemPtr->y1)*zoom/16);
		    XDestroyImage(xi);
		}
	} else if (itemPtr->kind == SLD_ITEM_TEXT) {
        Sld_DrawText(slidePtr,d,itemPtr,xoff,yoff);
	}
    if (itemPtr->kind != SLD_ITEM_TEXT)
        Tk_FreeGC(slidePtr->display, gc);
}

/*
 *--------------------------------------------------------------
 *
 * Sld_DeleteItem -- remove an item from the slide widget
 *
 *--------------------------------------------------------------
 */
int Sld_DeleteItem(slidePtr, argc, argv)
Slide *slidePtr;
int argc;
char **argv;
{
	SldItem *itemPtr, *prevItemPtr;
	SldLayer *layerPtr;
	int result, i, itemId;

	layerPtr = slidePtr->currentLayer;
	if (layerPtr == NULL) {
		sprintf(slidePtr->interp->result, "no current layer");
		return TCL_ERROR;
	}
	for (i = 0; i < argc; i++) {
		if (Tcl_GetInt(slidePtr->interp, argv[i], &itemId) != TCL_OK) {
			Tcl_AppendResult(slidePtr->interp, "wrong item id: ",
					 argv[i], (char *) NULL);
			return TCL_ERROR;
		}
		/*
		 * locate the item in list for current layer 
		 */
		itemPtr = layerPtr->firstItemPtr;
		prevItemPtr = NULL;
		while (itemPtr != NULL) {
			if (itemPtr->id == itemId)
				break;
			prevItemPtr = itemPtr;
			itemPtr = itemPtr->nextItem;
		}
		if (itemPtr == NULL) {
			sprintf(slidePtr->interp->result, "item %d not found", itemId);
			return TCL_ERROR;
		}
		if (prevItemPtr == NULL) {
			layerPtr->firstItemPtr = itemPtr->nextItem;
		} else {
			prevItemPtr->nextItem = itemPtr->nextItem;
		}
		if (layerPtr->lastItemPtr == itemPtr)
			layerPtr->lastItemPtr = prevItemPtr;

		if (itemPtr->kind == SLD_ITEM_IMAGE) {
			XDestroyImage(itemPtr->xi);
		}

		Tk_FreeFont(itemPtr->tkfont);
		/*Tk_FreeOptions( itemConfigSpecs, (char *)itemPtr, 
		   slidePtr->display, 0); */
		ckfree(itemPtr);
	}
	if (!((slidePtr->flags) & WINDOW_REDRAW)) {
		Tcl_DoWhenIdle(SlideDisplay, (ClientData) slidePtr);
		slidePtr->flags |= WINDOW_REDRAW;
	}
	return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Sld_UnglueItem -- unglue items of the current layer
 *
 *--------------------------------------------------------------
 */
int Sld_UnglueItem(slidePtr, argc, argv)
Slide *slidePtr;
int argc;
char **argv;
{
	SldItem *itemPtr, *prevItemPtr;
	SldLayer *layerPtr;
	int result, i, itemId;

	layerPtr = slidePtr->currentLayer;
	if (layerPtr == NULL) {
		sprintf(slidePtr->interp->result, "no current layer");
		return TCL_ERROR;
	}
	for (i = 0; i < argc; i++) {
		if (Tcl_GetInt(slidePtr->interp, argv[i], &itemId) != TCL_OK) {
			Tcl_AppendResult(slidePtr->interp, "wrong item id: ",
					 argv[i], (char *) NULL);
			return TCL_ERROR;
		}
		/*
		 * locate the item in list for current layer 
		 */
		itemPtr = layerPtr->firstGluedPtr;
		prevItemPtr = NULL;
		while (itemPtr != NULL) {
			if (itemPtr->id == itemId)
				break;
			prevItemPtr = itemPtr;
			itemPtr = itemPtr->nextItem;
		}
		if (itemPtr == NULL) {
			sprintf(slidePtr->interp->result, "item %d not found", itemId);
			return TCL_ERROR;
		}
		if (prevItemPtr == NULL) {
			layerPtr->firstGluedPtr = itemPtr->nextItem;
		} else {
			prevItemPtr->nextItem = itemPtr->nextItem;
		}
		if (layerPtr->lastGluedPtr == itemPtr)
			layerPtr->lastGluedPtr = prevItemPtr;

		itemPtr->nextItem = NULL;
		Sld_AppendItem(layerPtr, itemPtr);
	}
	if (!((slidePtr->flags) & BACKGROUND_RECONFIG))
		slidePtr->flags |= BACKGROUND_RECONFIG;
	return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Sld_ToLayerItem -- move items from the current to another layer
 *
 *--------------------------------------------------------------
 */
int Sld_ToLayerItem(slidePtr, argc, argv)
Slide *slidePtr;
int argc;
char **argv;
{
	SldItem *itemPtr, *prevItemPtr;
	SldLayer *layerPtr, *newLayerPtr;
	int result, i, itemId, layerId;

	layerPtr = slidePtr->currentLayer;
	if (layerPtr == NULL) {
		sprintf(slidePtr->interp->result, "no current layer");
		return TCL_ERROR;
	}
	if (Tcl_GetInt(slidePtr->interp, argv[0], &layerId) != TCL_OK) {
		Tcl_AppendElement(slidePtr->interp, "wrong layer id");
		return TCL_ERROR;
	}
	newLayerPtr = slidePtr->firstLayer;
	while (newLayerPtr != NULL) {
		if (newLayerPtr->id == layerId)
			break;
		newLayerPtr = newLayerPtr->nextLayer;
	}
	if (newLayerPtr == NULL) {
		Tcl_AppendElement(slidePtr->interp, "layer not found");
		return TCL_ERROR;
	}
	/*
	 * repeat the move-to-layer with all other parameters (item ids) 
	 */
	for (i = 1; i < argc; i++) {
		if (Tcl_GetInt(slidePtr->interp, argv[i], &itemId) != TCL_OK) {
			Tcl_AppendResult(slidePtr->interp, "wrong item id: ",
					 argv[i], (char *) NULL);
			return TCL_ERROR;
		}
		/*
		 * locate the item in list for current layer 
		 */
		itemPtr = layerPtr->firstItemPtr;
		prevItemPtr = NULL;
		while (itemPtr != NULL) {
			if (itemPtr->id == itemId)
				break;
			prevItemPtr = itemPtr;
			itemPtr = itemPtr->nextItem;
		}
		if (itemPtr == NULL) {
			sprintf(slidePtr->interp->result, "item %d not found", itemId);
			return TCL_ERROR;
		}
		if (prevItemPtr == NULL) {
			layerPtr->firstItemPtr = itemPtr->nextItem;
		} else {
			prevItemPtr->nextItem = itemPtr->nextItem;
		}
		if (layerPtr->lastItemPtr == itemPtr)
			layerPtr->lastItemPtr = prevItemPtr;

		/*
		 * insert this item in the layer requested 
		 */
		itemPtr->nextItem = NULL;
		if (newLayerPtr->lastItemPtr)
			newLayerPtr->lastItemPtr->nextItem = itemPtr;
		layerPtr->lastItemPtr = itemPtr;
		if (newLayerPtr->firstItemPtr == NULL)
			newLayerPtr->firstItemPtr = itemPtr;
	}
	if (!((slidePtr->flags) & WINDOW_REDRAW)) {
		Tcl_DoWhenIdle(SlideDisplay, (ClientData) slidePtr);
		slidePtr->flags |= WINDOW_REDRAW;
	}
	return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Sld_GlueItem -- glue items to the current layer
 *
 *--------------------------------------------------------------
 */
int Sld_GlueItem(slidePtr, argc, argv)
Slide *slidePtr;
int argc;
char **argv;
{
	SldItem *itemPtr, *prevItemPtr;
	SldLayer *layerPtr;
	int result, i, itemId;

	layerPtr = slidePtr->currentLayer;
	if (layerPtr == NULL) {
		sprintf(slidePtr->interp->result, "no current layer");
		return TCL_ERROR;
	}
	for (i = 0; i < argc; i++) {
		if (Tcl_GetInt(slidePtr->interp, argv[i], &itemId) != TCL_OK) {
			Tcl_AppendResult(slidePtr->interp, "wrong item id: ",
					 argv[i], (char *) NULL);
			return TCL_ERROR;
		}
		/*
		 * locate the item in list for current layer 
		 */
		itemPtr = layerPtr->firstItemPtr;
		prevItemPtr = NULL;
		while (itemPtr != NULL) {
			if (itemPtr->id == itemId)
				break;
			prevItemPtr = itemPtr;
			itemPtr = itemPtr->nextItem;
		}
		if (itemPtr == NULL) {
			sprintf(slidePtr->interp->result, "item %d not found", itemId);
			return TCL_ERROR;
		}
		if (prevItemPtr == NULL) {
			layerPtr->firstItemPtr = itemPtr->nextItem;
		} else {
			prevItemPtr->nextItem = itemPtr->nextItem;
		}
		if (layerPtr->lastItemPtr == itemPtr)
			layerPtr->lastItemPtr = prevItemPtr;

		/*
		 * put the item in the glued list 
		 */
		itemPtr->nextItem = NULL;
		if (layerPtr->lastGluedPtr)
			layerPtr->lastGluedPtr->nextItem = itemPtr;
		layerPtr->lastGluedPtr = itemPtr;
		if (layerPtr->firstGluedPtr == NULL)
			layerPtr->firstGluedPtr = itemPtr;
	}
	if (!((slidePtr->flags) & BACKGROUND_RECONFIG))
		slidePtr->flags |= BACKGROUND_RECONFIG;
	return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Sld_MoveItem -- move items
 *
 *--------------------------------------------------------------
 */
int Sld_MoveItem(slidePtr, argc, argv)
Slide *slidePtr;
int argc;
char **argv;
{
	SldItem *itemPtr, *prevItemPtr;
	SldLayer *layerPtr;
	int result, i, itemId;
	int deltax,deltay;

    if (argc < 3) {
        Tcl_AppendResult(slidePtr->interp,"wrong # args: must be",
            " delta_x delta_y itemId ?itemId?",NULL);
        return TCL_ERROR;
    }
    if (Tcl_GetInt(slidePtr->interp, argv[0], &deltax) != TCL_OK) {
    	Tcl_AppendResult(slidePtr->interp, "wrong delta_x: ",
    			 argv[0], (char *) NULL);
    	return TCL_ERROR;
    }
    if (Tcl_GetInt(slidePtr->interp, argv[1], &deltay) != TCL_OK) {
    	Tcl_AppendResult(slidePtr->interp, "wrong delta_y: ",
    			 argv[1], (char *) NULL);
    	return TCL_ERROR;
    }

	layerPtr = slidePtr->currentLayer;
	if (layerPtr == NULL) {
		sprintf(slidePtr->interp->result, "no current layer");
		return TCL_ERROR;
	}
	for (i = 2; i < argc; i++) {
		if (Tcl_GetInt(slidePtr->interp, argv[i], &itemId) != TCL_OK) {
			Tcl_AppendResult(slidePtr->interp, "wrong item id: ",
					 argv[i], (char *) NULL);
			return TCL_ERROR;
		}
		/*
		 * locate the item in list for current layer 
		 */
		itemPtr = layerPtr->firstItemPtr;
		prevItemPtr = NULL;
		while (itemPtr != NULL) {
			if (itemPtr->id == itemId)
				break;
			prevItemPtr = itemPtr;
			itemPtr = itemPtr->nextItem;
		}
		if (itemPtr == NULL) {
			sprintf(slidePtr->interp->result, "item %d not found", itemId);
			return TCL_ERROR;
		}
        switch (itemPtr->kind) {
        case SLD_ITEM_LINE:
        case SLD_ITEM_IMAGE:
        case SLD_ITEM_TEXT:
            itemPtr->x1 += deltax;
            itemPtr->y1 += deltay;
            itemPtr->x2 += deltax;
            itemPtr->y2 += deltay;
            break;
        case SLD_ITEM_RECTANGLE:
        case SLD_ITEM_ARC:
            itemPtr->x1 += deltax;
            itemPtr->y1 += deltay;
            break;
        }
    }
	return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Sld_RaiseItem -- place items in the top of current layer
 *
 *--------------------------------------------------------------
 */
int Sld_RaiseItem(slidePtr, argc, argv)
Slide *slidePtr;
int argc;
char **argv;
{
	SldItem *itemPtr, *prevItemPtr;
	SldLayer *layerPtr;
	int result, i, itemId;

	layerPtr = slidePtr->currentLayer;
	if (layerPtr == NULL) {
		sprintf(slidePtr->interp->result, "no current layer");
		return TCL_ERROR;
	}
	for (i = 0; i < argc; i++) {
		if (Tcl_GetInt(slidePtr->interp, argv[i], &itemId) != TCL_OK) {
			Tcl_AppendResult(slidePtr->interp, "wrong item id: ",
					 argv[i], (char *) NULL);
			return TCL_ERROR;
		}
		/*
		 * locate the item in list for current layer 
		 */
		itemPtr = layerPtr->firstItemPtr;
		prevItemPtr = NULL;
		while (itemPtr != NULL) {
			if (itemPtr->id == itemId)
				break;
			prevItemPtr = itemPtr;
			itemPtr = itemPtr->nextItem;
		}
		if (itemPtr == NULL) {
			sprintf(slidePtr->interp->result, "item %d not found", itemId);
			return TCL_ERROR;
		}
		if (prevItemPtr == NULL) {
			layerPtr->firstItemPtr = itemPtr->nextItem;
		} else {
			prevItemPtr->nextItem = itemPtr->nextItem;
		}
		if (layerPtr->lastItemPtr == itemPtr)
			layerPtr->lastItemPtr = prevItemPtr;

		/*
		 * put item back in the bottom of the list  so it will cover the
		 * others 
		 */
		itemPtr->nextItem = NULL;
		if (layerPtr->lastItemPtr)
			layerPtr->lastItemPtr->nextItem = itemPtr;
		layerPtr->lastItemPtr = itemPtr;
		if (layerPtr->firstItemPtr == NULL)
			layerPtr->firstItemPtr = itemPtr;
	}
	if (!((slidePtr->flags) & WINDOW_REDRAW)) {
		Tcl_DoWhenIdle(SlideDisplay, (ClientData) slidePtr);
		slidePtr->flags |= WINDOW_REDRAW;
	}
	return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Sld_LowerItem -- place items in the bottom of current layer
 *
 *--------------------------------------------------------------
 */
int Sld_LowerItem(slidePtr, argc, argv)
Slide *slidePtr;
int argc;
char **argv;
{
	SldItem *itemPtr, *prevItemPtr;
	SldLayer *layerPtr;
	int result, i, itemId;

	layerPtr = slidePtr->currentLayer;
	if (layerPtr == NULL) {
		sprintf(slidePtr->interp->result, "no current layer");
		return TCL_ERROR;
	}
	for (i = 0; i < argc; i++) {
		if (Tcl_GetInt(slidePtr->interp, argv[i], &itemId) != TCL_OK) {
			Tcl_AppendResult(slidePtr->interp, "wrong item id: ",
					 argv[i], (char *) NULL);
			return TCL_ERROR;
		}
		/*
		 * locate the item in list for current layer 
		 */
		itemPtr = layerPtr->firstItemPtr;
		prevItemPtr = NULL;
		while (itemPtr != NULL) {
			if (itemPtr->id == itemId)
				break;
			prevItemPtr = itemPtr;
			itemPtr = itemPtr->nextItem;
		}
		if (itemPtr == NULL) {
			sprintf(slidePtr->interp->result, "item %d not found", itemId);
			return TCL_ERROR;
		}
		if (prevItemPtr == NULL) {
			layerPtr->firstItemPtr = itemPtr->nextItem;
		} else {
			prevItemPtr->nextItem = itemPtr->nextItem;
		}
		if (layerPtr->lastItemPtr == itemPtr)
			layerPtr->lastItemPtr = prevItemPtr;

		/*
		 * put the item in the top of the list  so it will be drawn first
		 * than the others 
		 */
		itemPtr->nextItem = layerPtr->firstItemPtr;
		layerPtr->firstItemPtr = itemPtr;
		if (layerPtr->lastItemPtr == NULL)
			layerPtr->lastItemPtr = itemPtr;
	}
	if (!((slidePtr->flags) & WINDOW_REDRAW)) {
		Tcl_DoWhenIdle(SlideDisplay, (ClientData) slidePtr);
		slidePtr->flags |= WINDOW_REDRAW;
	}
	return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Sld_SelectItem -- get items containing the point (bounding box)
 *
 *--------------------------------------------------------------
 */
int Sld_SelectItem(slidePtr, argc, argv)
Slide *slidePtr;
int argc;
char **argv;
{
	SldItem *itemPtr, *prevItemPtr;
	SldLayer *layerPtr;
	int result, itemId, x, y;

	layerPtr = slidePtr->currentLayer;
	if (layerPtr == NULL) {
		sprintf(slidePtr->interp->result, "no current layer");
		return TCL_ERROR;
	}
	if (Tcl_GetInt(slidePtr->interp, argv[0], &x) != TCL_OK) {
		Tcl_AppendResult(slidePtr->interp, "wrong item coord: ",
				 argv[0], (char *) NULL);
	}
	if (Tcl_GetInt(slidePtr->interp, argv[1], &y) != TCL_OK) {
		Tcl_AppendResult(slidePtr->interp, "wrong item coord: ",
				 argv[1], (char *) NULL);
	}
    x = (x + slidePtr->xView)*16/slidePtr->zoom;
    y = (y + slidePtr->yView)*16/slidePtr->zoom;
    
	for (itemPtr = layerPtr->firstGluedPtr; itemPtr != NULL;
	     itemPtr = itemPtr->nextItem) {
		appendSelectElement(slidePtr->interp, itemPtr, x, y);
	}
	for (itemPtr = layerPtr->firstItemPtr; itemPtr != NULL;
	     itemPtr = itemPtr->nextItem) {
		appendSelectElement(slidePtr->interp, itemPtr, x, y);
	}
	return TCL_OK;
      error:
	return TCL_ERROR;
}

static void appendSelectElement(interp, itemPtr, x, y)
Tcl_Interp *interp;
SldItem *itemPtr;
int x, y;
{
	int bx1, by1, bx2, by2;
	char buff[10];

	if (itemPtr->x1 < itemPtr->x2) {
		bx1 = itemPtr->x1;
		bx2 = itemPtr->x2;
	} else {
		bx2 = itemPtr->x1;
		bx1 = itemPtr->x2;
	}
	if (itemPtr->y1 < itemPtr->y2) {
		by1 = itemPtr->y1;
		by2 = itemPtr->y2;
	} else {
		by2 = itemPtr->y1;
		by1 = itemPtr->y2;
	}
    bx1-=5; bx2+=5;
    by1-=5; by2+=5;

	if ((bx1 <= x) && (bx2 >= x) &&
	    (by1 <= y) && (by2 >= y)) {
		sprintf(buff, "%d", itemPtr->id);
		Tcl_AppendElement(interp, buff);
	}
}

/*
 *--------------------------------------------------------------
 *
 * Sld_Rubberband -- toggle rubberband drawing
 *
 *--------------------------------------------------------------
 */
int Sld_Rubberband(slidePtr, argc, argv)
Slide *slidePtr;
int argc;
char **argv;
{
	Tk_Window tkwin = slidePtr->tkwin;
	GC gc;
	XGCValues gcv;
	char c;
	int length;
	int rx1, ry1, rx2, ry2, kind;
	int w, h;

	c = argv[0][0];
	length = strlen(argv[0]);
	if ((c == 'l') && (strncmp(argv[0], "line", length) == 0)
	    && (length >= 2)) {
		kind = 1;
	} else if ((c == 'r') && (strncmp(argv[0], "rectangle", length) == 0)
		   && (length >= 2)) {
		kind = 2;
	} else if ((c == 'o') && (strncmp(argv[0], "oval", length) == 0)
		   && (length >= 2)) {
		kind = 3;
	} else {
		Tcl_AppendResult(slidePtr->interp, "bad option \"", argv[0],
		     "\": must be line, rectangle, or oval", (char *) NULL);
		goto error2;
	}

	if (Tcl_GetInt(slidePtr->interp, argv[1], &rx1) != TCL_OK) {
		goto error;
	}
	if (Tcl_GetInt(slidePtr->interp, argv[2], &ry1) != TCL_OK) {
		goto error;
	}
	if (Tcl_GetInt(slidePtr->interp, argv[3], &rx2) != TCL_OK) {
		goto error;
	}
	if (Tcl_GetInt(slidePtr->interp, argv[4], &ry2) != TCL_OK) {
		goto error;
	}
	
    rx1 = rx1*slidePtr->zoom/16-slidePtr->xView;
    ry1 = ry1*slidePtr->zoom/16-slidePtr->yView;
    rx2 = rx2*slidePtr->zoom/16-slidePtr->xView;
    ry2 = ry2*slidePtr->zoom/16-slidePtr->yView;
    	
	/* show rubberband */
	if (!Tk_IsMapped(tkwin)) {
		return;
	}
	if (kind) {
		if (kind > 1) {
			w = rx2 - rx1;
			h = ry2 - ry1;
			if (w < 0) {
				w = -w;
				rx1 = rx2;
			}
			if (h < 0) {
				h = -h;
				ry1 = ry2;
			}
		}
		if (w==0) w++;
		if (h==0) h++;
		gcv.foreground = WhitePixel(Tk_Display(tkwin), Tk_ScreenNumber(tkwin));
		gcv.function = GXxor;
		gc = Tk_GetGC(tkwin, GCForeground | GCFunction, &gcv);
		if (kind == 2) {
			XDrawRectangle(slidePtr->display, Tk_WindowId(tkwin), gc,
				       rx1, ry1, w, h);
		} else if (kind == 3) {
			XDrawArc(slidePtr->display, Tk_WindowId(tkwin), gc,
				 rx1, ry1, w, h, 0, 360 * 64);
		} else {
			XDrawLine(slidePtr->display, Tk_WindowId(tkwin), gc,
				  rx1, ry1, rx2, ry2);
		}
		Tk_FreeGC(Tk_Display(tkwin), gc);
	}
	return TCL_OK;
      error:
	Tcl_AppendResult(slidePtr->interp, "wrong # args: should be \"",
			 "rubberbandKind x1 y1 x2 y2\"", (char *) NULL);
      error2:
	slidePtr->rubberband = 0;
	return TCL_ERROR;
}

void Sld_SetScrollbars( slidePtr )
Slide *slidePtr;
{
	if (slidePtr->xScrollCmd) {
        char buffer[50];
        double fraction,fraction2;

        fraction = (double)slidePtr->xView/(double)slidePtr->sWidth;
        fraction2 = (double)(slidePtr->xView+slidePtr->width*
            16/slidePtr->zoom)/(double)slidePtr->sWidth;
        fraction = FORCE_LIMITS(fraction,0.0,1.0);
        fraction2 = FORCE_LIMITS(fraction2,0.0,1.0);
        sprintf(buffer,"%g %g",fraction,fraction2);
        if (Tcl_VarEval(slidePtr->interp,slidePtr->xScrollCmd, " ",
            buffer,NULL) != TCL_OK)
                Tcl_BackgroundError(slidePtr->interp);
        Tcl_ResetResult(slidePtr->interp);
	}	
	if (slidePtr->yScrollCmd) {
        char buffer[50];
        double fraction,fraction2;

        fraction = (double)slidePtr->yView/(double)slidePtr->sHeight;
        fraction2 = (double)(slidePtr->yView+slidePtr->height*
            16/slidePtr->zoom)/(double)slidePtr->sHeight;
        fraction = FORCE_LIMITS(fraction,0.0,1.0);
        fraction2 = FORCE_LIMITS(fraction2,0.0,1.0);
        sprintf(buffer,"%g %g",fraction,fraction2);
        if (Tcl_VarEval(slidePtr->interp,slidePtr->yScrollCmd, " ",
            buffer,NULL) != TCL_OK)
                Tcl_BackgroundError(slidePtr->interp);
        Tcl_ResetResult(slidePtr->interp);
	}
}
