/* 
 * slide widget definitions
 */
#define FORCE_LIMITS(x,xmin,xmax) (((x) < (xmin))?(xmin):((x) > (xmax))?(xmax):(x))
 
typedef struct SldItem {
    struct SldItem *nextItem;
    int id;		        /* identifier for this element */
    int kind;		    /* kind of element (line, image,...) */
    int x1,y1,x2,y2;	/* coordinates of the item */
    int angle1,angle2;  /* angles for arc drawing */
    int width;		    /* line width */
    XColor *color;      /* color for the item outline */
    XColor *fillColor;  /* color for filling the item */
    char *image;        /* filename for images, string for text */
    XImage *xi;         /* image storage */
    char *text;         /* string for text items */
    int rotate;         /* rotation angle */
    Tk_Font tkfont;     /* font for string */    
} SldItem;

#define SLD_ITEM_LINE 1
#define SLD_ITEM_RECTANGLE 2
#define SLD_ITEM_ARC 3
#define SLD_ITEM_OVAL 4
#define SLD_ITEM_IMAGE 5
#define SLD_ITEM_TEXT 6

#define SLD_LINE_MASK       TK_CONFIG_USER_BIT
#define SLD_RECTANGLE_MASK  TK_CONFIG_USER_BIT << 1
#define SLD_ARC_MASK        TK_CONFIG_USER_BIT << 2
#define SLD_OVAL_MASK       TK_CONFIG_USER_BIT << 3
#define SLD_IMAGE_MASK      TK_CONFIG_USER_BIT << 4
#define SLD_TEXT_MASK       TK_CONFIG_USER_BIT << 5
#define SLD_ALL_MASK (SLD_LINE_MASK | SLD_RECTANGLE_MASK | \
    SLD_ARC_MASK | SLD_OVAL_MASK | SLD_IMAGE_MASK | SLD_TEXT_MASK)
     
typedef struct SldLayer {
    int id;                     /* identifier for the layer */
    int visible;                /* if the layer is visible */
    int transparent;            /* if the layer is transparent or opaque */
    struct SldLayer *nextLayer;
    int nextId;                 /* id of next item created in this layer */
    SldItem *firstGluedPtr;      /* First glued item */
    SldItem *lastGluedPtr;       /* Last glued item */
    SldItem *firstItemPtr;      /* First in list of all items in layer */
    SldItem *lastItemPtr;       /* Last in list of all items in layer */
} SldLayer;

typedef struct {
    Tk_Window tkwin;		/* Window that embodies the slide.  NULL
				            * means window has been deleted but
				            * widget record hasn't been cleaned up yet. */
    Display *display;		/* X's token for the window's display. */
    Tcl_Interp *interp;		/* Interpreter associated with widget. */
    Tcl_Command widgetCmd;	/* Token for slide's widget command. */

    SldLayer *currentLayer;
    SldLayer *firstLayer;
    SldLayer *lastLayer;
    int nextId;             /* id of next layer to create */
    Window wshow;           /* window for doing slide show */
    
    /*
     * Information used when displaying widget:
     */

    int rubberband;         /* 0=don't draw, 1=line, 2=rectangle, 3=oval */
    int rx1,ry1,rx2,ry2;    /* rubberband limits */
    int zoom;			    /* zoom level for displaying */
    Pixmap backPixmap;      /* background pixmap */
    int grid;               /* grid spacing (0=disable) */
    int gridsize;           /* size of grid points */
    int gridNpoints;        /* number of points in all grid */
    XPoint *gridPoints;     /* array of grid points */
    XColor *gridColor;      /* color for grid points */
    char *gradient;		    /* pattern for the background, if any */
    XColor *gradbegin;	    /* begin color of gradients */
    XColor *gradend;	    /* end color of gradients */
    XColor *outcolor;       /* color outside slide drawing region */
    int borderWidth;		/* Width of 3-D border around whole widget. */
    int width,height;		/* size of the outer window of the widget */
    int sWidth,sHeight;     /* total size of the slide */
    int xView,yView;        /* visible area top left */
    char *xScrollCmd;       /* command to set scrollbar horizontal */
    char *yScrollCmd;       /* command to set scrollbar vertical */
    Tk_3DBorder bgBorder;	/* Used for drawing background. */
    Tk_3DBorder fgBorder;	/* For drawing slide. */
    int relief;			    /* Indicates whether window as a whole is
				            * raised, sunken, or flat. */
    GC gc;			        /* Graphics context for copying from
				            * off-screen pixmap onto screen. */
    int flags;
} Slide;

/*** flags definitions ***/
#define BACKGROUND_RECONFIG 1
#define GRID_RECONFIG 2
#define RUBBERBAND_REDRAW 4
#define WINDOW_REDRAW 8
#define CHANGE_ZOOM 16
#define UPDATE_SCROLLBARS 32

/* slide.c */
int SlideCmd(ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
char *Sld_MakeGradientBitmap(int width, int height);
int Sld_DeleteLayers(Slide *slidePtr, int argc, char **argv);
int Sld_LayerCmd(Slide *slidePtr, int argc, char **argv);
void Sld_DisplayLayer(Slide *slidePtr, Drawable d, SldLayer *layerPtr);
void Sld_DisplayGluedLayer(Slide *slidePtr, Drawable d, SldLayer *layerPtr);
void Sld_AppendItem(SldLayer *layerPtr, SldItem *itemPtr);
int Sld_CreateItem(Slide *slidePtr, int argc, char **argv);
int Sld_ConfigureItem(Slide *slidePtr, int itemId, int argc, char **argv);
int Sld_ItemCoords(Slide *slidePtr, int itemId, int argc, char **argv);
void Sld_DisplayItem(Slide *slidePtr, Drawable d, SldItem *itemPtr, 
    int xoff, int yoff);
int Sld_DeleteItem(Slide *slidePtr, int argc, char **argv);
int Sld_UnglueItem(Slide *slidePtr, int argc, char **argv);
int Sld_ToLayerItem(Slide *slidePtr, int argc, char **argv);
int Sld_GlueItem(Slide *slidePtr, int argc, char **argv);
int Sld_RaiseItem(Slide *slidePtr, int argc, char **argv);
int Sld_LowerItem(Slide *slidePtr, int argc, char **argv);
int Sld_SelectItem(Slide *slidePtr, int argc, char **argv);
int Sld_Rubberband(Slide *slidePtr, int argc, char **argv);
void Sld_SetScrollbars( Slide *slidePtr );

/* jpeg.c */
unsigned char *Sld_read_jpeg(char *filename, int *width, int *height);
void Sld_write_jpeg(char *data, char *filename, int width, int height);

/* sldPs.c */
int Sld_ps_output( Slide *slidePtr, char *filename );

/* sldImage.c */
int Sld_alloc_colormap(Tk_Window tkwin);
unsigned long Sld_pixel16(int red, int green, int blue);
int Sld_loadjpg(Slide *slidePtr, SldItem *itemPtr);
int Sld_savejpeg( Slide *slidePtr, char *filename );
XImage *Sld_zoom_image( Tk_Window tkwin, XImage *xi_orig, int zoom);
void Sld_DrawBackground( Slide *slidePtr, int zoom, Pixmap d1 );
int Sld_show( Slide *slidePtr, char *transisition );
